/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : fcover.cc                                                */
/*                                                                       */
/*       Date : March     04, 1998                                       */
/* Updated at : September 15, 1998                                       */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/**
    Finds a solution for Set Cover Problem by using various methods.

    - GREEDY
      * CHAVTAL
      * BACK
      * GREEDY3
    - RANDOM
    - BCGA Beasley&Cgu GA
    - SSGA with various operator combinations.

*/
/*************************************************************************/

#include <time.h>
#include <limits.h>
#include <iomanip.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <new.h>

#include <Population.h>
#include <Sort_Recombinator.h>
#include <Stat_Recombinator.h>
#include <SimilarParentSelector.h>
#include <GeneticAlgorithm.h>

#include "File1Reader.h"
#include "SetCoverProblem.h"
#include "SetCoverSolution.h"
#include "SCPFitnessFunction.h"

#include "SCPMutation.h"
#include "SCPMutationRnd.h"
#include "SCPMutationBack.h"
#include "SCPMutationChavatal.h"
#include "SCPMutationGreedy3.h"

#include "SCPCrossover.h"
#include "SCPCrossoverFusion.h"
#include "SCPCrossoverDPXRnd.h"
#include "SCPCrossoverDPXBack.h"
#include "SCPCrossoverDPXChavatal.h"
#include "SCPCrossoverDPXGreedy3.h"
#include "SCPCrossover1PX.h"
#include "SCPCrossover1PXBack.h"
#include "SCPCrossover1PXChv.h"
#include "SCPCrossover1PXGr3.h"
#include "SCPCrossover2PX.h"
#include "SCPCrossover2PXBack.h"
#include "SCPCrossover2PXChv.h"
#include "SCPCrossover2PXGr3.h"
#include "SCPCrossoverUniform.h"
#include "SCPCrossoverUniformBack.h"
#include "SCPCrossoverUniformChv.h"
#include "SCPCrossoverUniformGr3.h"
#include "SCPInitializer.h"

#include "SCPChromosomes.h"
#include "SCPTestDefinition.h"

#define   PopSize   20
#define   GenNum   100

//typedef   long time_t;

/*************************************************************************/
/*************************************************************************/

void allocation_handler()
{
   cout << "Out of Memory. Aborting Program...."<< endl;
   exit(1);
}

/*************************************************************************/
/*************************************************************************/

void printReport(unsigned elapsedTime, ostream& out, SetCoverSolution* sol)
{
   out << "Cost ............: " << sol->getCost() << endl;
   out << "Number Columns ..: " << sol->getNumOfColumns() << endl;
   out << "Number Overlaps .: " << sol->getNumberOfOverlaps() << endl;
   out << "Time ............: " << elapsedTime << endl;
   SetCoverProblem *problem = sol->getProblem();

   if( problem!=NULL )
   {
      out << "Problem " << endl;

      out << "  Number Columns .........: " << problem->getNumOfCols() << endl;

      out << "  Number Rows ............: " << problem->getNumOfRows() << endl;

      out << "  Ave Num of Col per row .: " 
	  << problem->getAverageNumOfColumns() << endl;

      out << "  Ave Num of Row per col .: " 
	  << problem->getAverageNumOfRows() << endl;

      out << "  Density ................: " << problem->getDensity() << endl;

   }// end of if

}// end of printReport()

/*************************************************************************/
/*************************************************************************/

void printReport(char *repFileName, unsigned elapsedTime, SetCoverSolution* sol)
{
   ofstream out(repFileName);
   printReport(elapsedTime,out,sol);
   out.close(); 

}// end of printReport()

/*************************************************************************/
/*************************************************************************/

void printSolution(char *solFileName, SetCoverSolution* sol)
{
   ofstream f1(solFileName);
   sol->printColumnsBinary(f1);
   f1.close(); 

}// end of printSolution()

/*************************************************************************/
/**
     Prints:
     1: Average Fitness Value
     2: Best Chromosomes:
	2.a. Fitness Value
	2.b. Total Cost
	2.c. Number of Columns
	2.d. NUmber of overlaps
     3: Worst Chromosomes 
	3.a. Fitness Value
	3.b. Total Cost
	3.c. Number of Columns
	3.d. NUmber of overlaps
*/
/*************************************************************************/

void printStat(char* fileName, Population& Pop)
{
   ofstream out(fileName, ios::app); 

   FitnessValue *fv = Pop.getAverageFitness();
   unsigned *srt    = Pop.getSortIndex(); 

   out << *fv << " "; // 1: Average Fitness Value
   delete fv;

   SCPChromosomes *a = (SCPChromosomes *)Pop.getChromosomesIndex(srt[0]);

   fv = a->getFitnessValue();
   out << *fv 
       << " " << a->getCost() 
       << " " << a->getNumOfColumns() 
       << " " << a->getNumberOfOverlaps() ;
   delete fv;

   a = (SCPChromosomes *)Pop.getChromosomesIndex(srt[Pop.getIndex()-1]);

   fv = a->getFitnessValue();
   out << *fv 
       << " " << a->getCost() 
       << " " << a->getNumOfColumns() 
       << " " << a->getNumberOfOverlaps() << endl;
   delete fv;

   delete[] srt;

   out.close();

}// end of printStat

/*************************************************************************/
/**
    We need a configuration file to pick several operator combinations
    for us.

    MutationType  { MutationBack, MutationRnd, MutationChavatal,
                    MutationGreedy3, MutationRandomDelete,
		    MutationEqualizer }

    CrossoverType { XPRnd, XPBack, XPChavatal, XPGreedy3,
		    DXPRnd, DXPBack, DXPChavatal, DXPGreedy3 }

    InitialFileName : Writes the inital solutions to this file

    ResultFileName  : Writes the final solutions to this file

    BestSolutionReportFileName : Writes the report of best solution

    BestSolutionBinaryOutputFileName : Writes the best solution 

*/
/*************************************************************************/

void solvewithGASTDptr( SetCoverProblem *SCPInstance, char *gaFile, 
		        char *RgaFile)
{
   SCPFitnessFunction   fitFunc;
   SCPInitializer       Initme(SCPInstance);
   //SCPMutation          Jmutation;
   //SCPMutationBack          Jmutation;
   SCPMutationRnd          Jmutation;
   //SCPCrossover         Jcross;
   SCPCrossoverDPXGreedy3         Jcross;
   Population           myPop(PopSize);    
   long       bias=0; 

   time_t start_t;
   time_t stop_t;
   time_t dummy;

   cout << " before init() " << endl;
   start_t = time(&dummy);
   Initme.apply(myPop);
   stop_t = time(&dummy);
   bias = stop_t-start_t;
   cout << " after init() took " << bias << endl;

   //Sort_Recombinator     SRecombinator;
   Stat_Recombinator     SRecombinator;
   SimilarParentSelector SimilarSelect;
    
   GeneticAlgorithm *mySolver = new GeneticAlgorithm(&fitFunc,&Jcross,&Jmutation,&SimilarSelect,&SRecombinator,&Initme);

   int MaxGenNum= GenNum;
   int PopulationSize = PopSize;
   float MutationProbability = 0.2;
   float CrossoverProbability = 0.6;
   int TestTimes = 1;
   int TestType = 0;
   int g;

   mySolver->SetMaxGeneration(MaxGenNum);
   mySolver->SetPopulationSize(PopulationSize);
   mySolver->SetMutationProb(MutationProbability);
   mySolver->SetCrossProb(CrossoverProbability);

   //mySolver.run();

   if(TestType>0&&TestType<4) mySolver->setTestType(TestType);

   cout << " before run() " << endl;

   //cout << " Current population " << myPop << endl << "0000" << endl;
   Population *nPop ;

   //ofstream oStat2("SCPSolution.ini"); 
   //oStat2 << " Initial Population " << endl << myPop << endl;
   //oStat2.close();

   
   //ofstream oStat3("GASolution.stat"); 
   //oStat3 << MaxGenNum << " " << PopulationSize <<" " << endl;
   //oStat3.close();

   char fileName[30];
   sprintf(fileName,"SCPtatisticsT.out");
   ofstream oStat(fileName); oStat.close();

   start_t = time(&dummy);
   for(int i=0;i<MaxGenNum;i++)
   {
      //writeStat(&myPop);
      cout << endl << endl << " GENERATION # " << i << endl << endl;
      nPop = mySolver->runOnceOn(myPop);
      //nPop->writeStatistics(i,fileName);
      myPop = SRecombinator.apply(myPop,*nPop);

      /* Static Recom
      if(nPop->getIndex())
      {
	 cout << " In Recombinator " << endl;
	 unsigned* sIndex1 = nPop->getSortIndex();
	 unsigned* sIndex2 = myPop.getSortIndex();

	 Chromosomes* chr = nPop->getChromosomesIndex(sIndex1[0]);
	 Chromosomes* chr2 = myPop.getChromosomesIndex(sIndex2[myPop.getIndex()-1]);
         if( chr!=NULL && chr2!=NULL && 
	     *(chr->getFitnessValue())>*(chr2->getFitnessValue()))
	 {
	    nPop->clearExceptIndx(sIndex1[0]);

            for(g=0;g<myPop.getIndex()-1;g++)
	    {
	        chr2 = myPop.getChromosomesIndex(sIndex2[g]);
		if( chr2!=NULL ) nPop->addChromosomes(chr2);

	    }// end of for

	    myPop = *nPop ;

	 }// end of if

	 delete[] sIndex1;
	 delete[] sIndex2;

      }// end of if

      */

      if(myPop.getSumOfDissimilarities()<0.1)
      {
         int m, nOfreplace = maximum(1,0.7*PopulationSize);
         unsigned* sIndex = myPop.getSortIndex();
         FitnessValue *fv; 

	 /*
         ofstream oStatT("GASolution.sort",ios::app); 
         oStatT << "  After " << i << " epoch " << endl ;
         for(m=0;m<PopulationSize;m++) 
	 {
            Chromosomes* chr = myPop.getChromosomesIndex(sIndex[m]);
            if(chr!=NULL && (fv=chr->getFitnessValue())!=NULL) 
            {  
	       //oStatT << "[ "<<m<<"] : " << *fv << endl; 
	       delete fv; 
	    }// end of if

         }// end of for

         oStatT.close();
	 */

         for(m=PopulationSize-nOfreplace-1;m<PopulationSize;m++)
	 {
             Chromosomes* chr = myPop.getChromosomesIndex(sIndex[m]);
             if(chr!=NULL) chr->randomInitialize();

         }// end of for


         delete[] sIndex;

      }// end of if

      printStat(fileName,myPop);

   }// end of for

   stop_t = time(&dummy);

   //cout << " Before  runOnceOn() " << endl;
   //nPop = mySolver->runOnceOn(&myPop,MaxGenNum);
   //myPop = *nPop;

   cout << " After one epoch " << myPop << endl;
   //nPop->printFitnessValues(cout);
   //mySolver.writeStatistics(i,fileName);

   cout << " Before return of solvewithGASTD() " << endl;

   ofstream oStat1("SCPSolution.out"); 
   oStat1 << " After " << MaxGenNum << " epoch " << endl << myPop << endl;
   oStat1 << " Time " << (stop_t-start_t) << " msec " << endl;
   oStat1.close();


   unsigned *srtIndex = myPop.getSortIndex();

   Chromosomes *best = myPop.getChromosomesIndex(srtIndex[0]);

   //ofstream out2("ga_b.out"); 
   //((SCPChromosomes*)best)->printColumnsBinary(out2);
   //out2.close();

   printReport(RgaFile,bias+(stop_t-start_t),((SCPChromosomes*)best));
   printSolution(gaFile,((SCPChromosomes*)best));

   delete[] srtIndex;

   return;

}// end of solvewithGASTDptr()

/*************************************************************************/
/*************************************************************************/
void printPopArray(Chromosomes **pop, int limit)
{
   float curr;
   int i;

   for(i=0;i<limit;i++)
   {
      cout << i << " " ;
      if( pop[i]!=NULL )
      {
         curr = 1.0/((SCPChromosomes*)pop[i])->getCost();
         cout << ((SCPChromosomes*)pop[i])->getCost() << " " << curr;

      }// end of if
      cout << endl;
      
   }//end of for

}// end of printPopArray()

/*************************************************************************/
/**
    Initialization for BCGA
    let limit denote the population size
    let P and size(P) denote the current population and size of the
        population respectively.
    let P = {}
    while( size(P) < limit )
    {
       find a solution (Si) with random choice
       apply reduction 
       P = P U Si  // note that there is no equality check
    }
*/
/*************************************************************************/
void InitializeBCGA(Chromosomes **pop, SetCoverProblem *SCPInstance, int limit)
{
   int i=0, k, j, shift=0;
   float curr;
   float currj;
   Chromosomes* dmy;

   for(i=0;i<limit;i++)
   {
      pop[i] = new SCPChromosomes(SCPInstance);
      ((SCPChromosomes*)pop[i])->solveWithRandomChoice();
      ((SCPChromosomes*)pop[i])->reduce();

      //cout << *pop[i] << endl << "i/limit: " << i << "/" << limit << endl;
      curr = 1.0/((SCPChromosomes*)pop[i])->getCost();

      shift=0;
      j=0;
      while( j<i && !shift )
      {
         currj = 1.0/((SCPChromosomes*)pop[j])->getCost();

         if( currj<curr ) shift=1;
	 else j++;

      }// end of while

      if(shift)
      {
	 //cout << " Shift necessary from " << j << " to " << i << endl;
	 dmy = pop[i];
         for(k=i;k>j;k--) pop[k]=pop[k-1];
	 pop[k] = dmy;

	  //printPopArray(pop,i);

       }// end of if

   }// end of for

   cout << " Population is ready" << endl;

   //printPopArray(pop,limit);

}// end of InitializeBCGA()

/*************************************************************************/
/*************************************************************************/

void writeStatForBCGA( ofstream& out, float *popFV, 
		       Chromosomes** pop, int gennum, int psize)
{
   float a=0;
   for(int i=0;i<psize;i++) a += ((SCPChromosomes*)pop[i])->getCost();
   out << gennum << " " 
       << ((SCPChromosomes*)pop[0])->getCost() << " "
       << ((SCPChromosomes*)pop[psize-1])->getCost() << " "
       << 1.0*a/psize << endl;

}// end of printReport()

/*************************************************************************/
/*************************************************************************/

void writeStatForBCGA( char *repFileName, float *popFV, 
		       Chromosomes** pop, int gennum, int psize)
{
   ofstream out(repFileName,ios::app);
   writeStatForBCGA(out,popFV,pop,gennum,psize);
   out.close(); 

}// end of writeStatForBCGA()

/*************************************************************************/
/**
*/
/*************************************************************************/
void reInitialize(Chromosomes **pop, float* popFV, int limit)
{
   int i=0, k, j, shift=0;
   float curr;
   float currj;
   Chromosomes* dmy;
   int start = 0.1*limit;
   if(start<1) start=1;

   for(i=start;i<limit;i++)
   {
      ((SCPChromosomes*)pop[i])->solveWithRandomChoice();
      ((SCPChromosomes*)pop[i])->reduce();

      //cout << *pop[i] << endl << "i/limit: " << i << "/" << limit << endl;
      curr = 1.0/((SCPChromosomes*)pop[i])->getCost();

      shift=0;
      j=0;
      while( j<i && !shift )
      {
         currj = 1.0/((SCPChromosomes*)pop[j])->getCost();

         if( currj<curr ) shift=1;
	 else j++;

      }// end of while

      if(shift)
      {
	 //cout << " Shift necessary from " << j << " to " << i << endl;
	 dmy = pop[i];
         for(k=i;k>j;k--) pop[k]=pop[k-1];
	 pop[k] = dmy;

	  //printPopArray(pop,i);

       }// end of if

   }// end of for

   cout << " Population is ready" << endl;

   //printPopArray(pop,limit);

}// end of reInitialize()

/*************************************************************************/
/**
    Applies Beasley&Chu GA Approach.

    We need a configuration file to pick several operator combinations
    for us.

    MutationType  { MutationBack, MutationRnd, MutationChavatal,
                    MutationGreedy3, MutationRandomDelete,
		    MutationEqualizer }

    CrossoverType { XPRnd, XPBack, XPChavatal, XPGreedy3,
		    DXPRnd, DXPBack, DXPChavatal, DXPGreedy3 }

    InitialFileName : Writes the inital solutions to this file

    ResultFileName  : Writes the final solutions to this file

    BestSolutionReportFileName : Writes the report of best solution

    BestSolutionBinaryOutputFileName : Writes the best solution 

*/
/*************************************************************************/

void solvewithGASTDptrBC( SetCoverProblem *SCPInstance, SCPTestDefinition test)
{
   SCPFitnessFunction   fitFunc;
   SCPInitializer       Initme(SCPInstance);
   SCPMutationBack      Jmutation;
   SCPCrossoverFusion   Jcross;
   long                 bias=0; 
   Chromosomes**        pop = NULL;
   float*               popFV = NULL;
   int MaxGenNum      = test.getNumberOfGenerations();
   int PopulationSize = test.getPopulationSize();
   float MutationProbability  = test.getMutationProbability();
   float CrossoverProbability = test.getCrossProbability();
   int g, i, k, j, duplicate, checkShift;
   time_t start_t;
   time_t stop_t;
   time_t dummy;
   char *gaFile = test.getBinResultFileName();
   char *RgaFile = test.getReportFileName();
   int  pa1Indx, pa2Indx;
   ChromosomesBag ChroBag;
   Chromosomes*   candidate;
   float  candidFV;
   float  currj;

   // Define Population
   pop = new Chromosomes*[PopulationSize];
   popFV = new float[PopulationSize];
   for(i=0;i<PopulationSize;i++) { pop[i]=NULL; popFV[i]=FLT_MAX; }

   // Initialize Population
   cout << " before init() " << endl;
   start_t = time(&dummy);
   InitializeBCGA(pop,SCPInstance,PopulationSize);
   stop_t = time(&dummy);
   bias = stop_t-start_t;

   SCPChromosomes* offspring = new SCPChromosomes(SCPInstance); 

   // Fill FitnessValue Array
   for(i=0;i<PopulationSize;i++) popFV[i] = 1.0/((SCPChromosomes*)pop[i])->getCost(); 

   cout << " after init() took " << bias << endl;

   float a, tot=0;
   for(i=0;i<PopulationSize;i++)  tot += popFV[i]; //->getLinearValue(); 
   float ftot = tot;
   float mul ; //= tot/50.0;

   char fileName[30];
   sprintf(fileName,"%s",test.getStatFileName());
   ofstream oStat(fileName); oStat.close();

   int cnt=0;
   int prev_i;
   start_t = time(&dummy);
   prev_i=i=0;
   while(i<MaxGenNum)
   {
      // ftot is the sum of fitness values
      mul = ftot / 50.0;

      // select first parent
      a = mul * (rand()%101);
      while(a>ftot) a-=ftot;
      tot = 0;k=0;
      while( tot<a ) 
      { tot += popFV[k%PopulationSize]; k++; }
      pa2Indx=pa1Indx=k%PopulationSize;

      // select second parent
      while( pa2Indx==pa1Indx )
      {
         a = mul * (rand()%101);
         while(a>ftot) a-=ftot;
         tot = 0;k=0;
         while( tot<a ) 
         { tot += popFV[k%PopulationSize]; k++; }

         pa2Indx=k%PopulationSize;

      }// end of while

      // Now we know two parents

      cout << "apply crossover between " << pa1Indx 
	   << " and " << pa2Indx  << " " 
	   << i << "/" << MaxGenNum << endl;

      //ChroBag = Jcross.apply(pop[pa1Indx],pop[pa2Indx]);

      //if(ChroBag.getNumberOfElements())
      //{
	 //candidate = ChroBag.getElementIndex(0);

	 offspring->clearAll();
         Jcross.applyf(pop[pa1Indx],pop[pa2Indx],offspring);
	 candidate = offspring; //ChroBag.getElementIndex(0);

	 candidate = Jmutation.apply(candidate);

         checkShift = duplicate = 0; 
         
	 for(k=0;k<PopulationSize&&!duplicate;k++) 
	     duplicate = (*candidate==*pop[k]);

         // Make sure that it is not duplicate
	 if(!duplicate)
	 {

	    //cout << " It is not duplicate!" << endl;
            candidFV = 1.0/((SCPChromosomes*)candidate)->getCost();

	    // If it is better, bubble up is necessary
	    if( candidFV>popFV[PopulationSize-1] ) checkShift=1;

	    // delete the last element
	    //delete pop[PopulationSize-1];
	    offspring=(SCPChromosomes*)pop[PopulationSize-1];

	    // Update the sum of fitness values
	    ftot = ftot - popFV[PopulationSize-1] + candidFV;

	    k=PopulationSize-1;
	    if( checkShift )
	    {
	        cout << " Shift needed " << endl;
		j=0;
		while( j<k && popFV[j]>candidFV) j++;

	        cout << " Shift needed " << j << endl;

		for(int m=k;m>j;m--) 
		{
		    pop[m] = pop[m-1];
		    popFV[m] = popFV[m-1];
		}
		k=m;
	    }// end of if

            // Add the new chromosome into the population
	    pop[k] = candidate;
	    popFV[k] = candidFV;
            //printPopArray(pop,PopulationSize);

            // Print some statistics
	    if( (i%10)==0) writeStatForBCGA(fileName,popFV,pop,i,PopulationSize);
	    i++;
	 }
	 //else delete candidate;

      //}// end of if

      if( prev_i==i ) cnt++;
      else cnt=0;

      // If it is not proucing anything, then restart some of the population
      if(cnt>20) { reInitialize(pop,popFV,PopulationSize); cnt=0; }
      prev_i=i;

   }// end of while
   stop_t = time(&dummy);
   writeStatForBCGA(fileName,popFV,pop,i,PopulationSize);

   printReport(RgaFile,bias+(stop_t-start_t),((SCPChromosomes*)pop[0]));
   printSolution(gaFile,((SCPChromosomes*)pop[0]));

   // Reclaim the memory 
   delete[] popFV;
   for(i=0;i<PopulationSize;i++)  delete pop[i];
   delete[] pop;

   return;

}// end of solvewithGASTDptr()

/*************************************************************************/
/**
    Applies Steady State GA Approach.

    We need a configuration file to pick several operator combinations
    for us.

    MutationType  { MutationBack, MutationRnd, MutationChavatal,
                    MutationGreedy3, MutationRandomDelete,
		    MutationEqualizer }

    CrossoverType { XPRnd, XPBack, XPChavatal, XPGreedy3,
		    DXPRnd, DXPBack, DXPChavatal, DXPGreedy3 }

    InitialFileName : Writes the inital solutions to this file

    ResultFileName  : Writes the final solutions to this file

    BestSolutionReportFileName : Writes the report of best solution

    BestSolutionBinaryOutputFileName : Writes the best solution 

*/
/*************************************************************************/

void solvewithSSGA( SetCoverProblem *SCPInstance, SCPTestDefinition test)
{
   SCPFitnessFunction   fitFunc;
   SCPInitializer       Initme(SCPInstance);
   MutationOperator     *Jmutation = NULL; //new SCPMutationBack();
   CrossOverOperation   *Jcross    = NULL; //new SCPCrossoverFusion();
   long                 bias=0; 
   Chromosomes**        pop = NULL;
   float*               popFV = NULL;
   int MaxGenNum      = test.getNumberOfGenerations();
   int PopulationSize = test.getPopulationSize();
   float MutationProbability  = test.getMutationProbability();
   float CrossoverProbability = test.getCrossProbability();
   int g, i, k, j, duplicate, checkShift;
   time_t start_t;
   time_t stop_t;
   time_t dummy;
   char *gaFile = test.getBinResultFileName();
   char *RgaFile = test.getReportFileName();
   int  pa1Indx, pa2Indx;
   ChromosomesBag ChroBag;
   Chromosomes*   candidate;
   float  candidFV;
   float  currj;

   // Define Population
   pop = new Chromosomes*[PopulationSize];
   popFV = new float[PopulationSize];
   for(i=0;i<PopulationSize;i++) { pop[i]=NULL; popFV[i]=FLT_MAX; }

   // Initialize Population
   cout << " before init() " << endl;
   start_t = time(&dummy);
   InitializeBCGA(pop,SCPInstance,PopulationSize);
   stop_t = time(&dummy);
   bias = stop_t-start_t;

   // Fill FitnessValue Array
   for(i=0;i<PopulationSize;i++) 
       popFV[i] = 1.0/((SCPChromosomes*)pop[i])->getCost(); 

   cout << " after init() took " << bias << endl;

   float a, tot=0;
   for(i=0;i<PopulationSize;i++)  tot += popFV[i]; //->getLinearValue(); 
   float ftot = tot;
   float mul ; //= tot/50.0;

   char fileName[30];
   sprintf(fileName,"%s",test.getStatFileName());
   ofstream oStat(fileName); oStat.close();

   // Decide about the Crossover Operator
   switch(test.getCrossoverOprType())
   {
      case SCPTestDefinition::RND1PX : 
           Jcross    = new SCPCrossover1PX();
           break;
      case SCPTestDefinition::CHV1PX : 
           Jcross    = new SCPCrossover1PXChv();
           break;
      case SCPTestDefinition::BCK1PX : 
           Jcross    = new SCPCrossover1PXBack();
           break;
      case SCPTestDefinition::GR31PX : 
           Jcross    = new SCPCrossover1PXGr3();
           break;
      case SCPTestDefinition::RND2PX : 
           Jcross    = new SCPCrossover2PX();
           break;
      case SCPTestDefinition::CHV2PX : 
           Jcross    = new SCPCrossover2PXChv();
           break;
      case SCPTestDefinition::BCK2PX : 
           Jcross    = new SCPCrossover2PXBack();
           break;
      case SCPTestDefinition::GR32PX : 
           Jcross    = new SCPCrossover2PXGr3();
           break;
      case SCPTestDefinition::UNIRND : 
           Jcross    = new SCPCrossoverUniform();
           break;
      case SCPTestDefinition::UNICHV : 
           Jcross    = new SCPCrossoverUniformChv();
           break;
      case SCPTestDefinition::UNIBCK : 
           Jcross    = new SCPCrossoverUniformBack();
           break;
      case SCPTestDefinition::UNIGR3 : 
           Jcross    = new SCPCrossoverUniformGr3();
           break;
      case SCPTestDefinition::XDPRND : 
           Jcross    = new SCPCrossoverDPXRnd();
           break;
      case SCPTestDefinition::XDPCHV : 
           Jcross    = new SCPCrossoverDPXChavatal();
           break;
      case SCPTestDefinition::XDPBCK : 
           Jcross    = new SCPCrossoverDPXBack();
           break;
      default: //case SCPTestDefinition::FUSION : 
           Jcross    = new SCPCrossoverFusion();
           break;
   }// end of switch()

   // Decide about the Mutation Operator
   switch(test.getMutationOprType())
   {
      case SCPTestDefinition::DPXRND : 
           Jmutation    = new SCPMutationRnd();
           break;
      case SCPTestDefinition::DPXCHV : 
           Jmutation    = new SCPMutationChavatal();
           break;
      case SCPTestDefinition::DPXGR3 : 
           Jmutation    = new SCPMutationGreedy3();
           break;
      default: //case SCPTestDefinition::DPXBCK : 
           Jmutation    = new SCPMutationBack();
           break;
   }// end of switch()

   int cnt=0;
   int prev_i;
   start_t = time(&dummy);
   prev_i=i=0;
   while(i<MaxGenNum)
   {
      // ftot is the sum of fitness values
      mul = ftot / 50.0;

      // select first parent
      a = mul * (rand()%101);
      while(a>ftot) a-=ftot;
      tot = 0;k=0;
      while( tot<a ) 
      { tot += popFV[k%PopulationSize]; k++; }
      pa2Indx=pa1Indx=k%PopulationSize;

      // select second parent
      while( pa2Indx==pa1Indx )
      {
         a = mul * (rand()%101);
         while(a>ftot) a-=ftot;
         tot = 0;k=0;
         while( tot<a ) 
         { tot += popFV[k%PopulationSize]; k++; }

         pa2Indx=k%PopulationSize;

      }// end of while

      // Now we know two parents

      cout << "apply crossover between " << pa1Indx 
	   << " and " << pa2Indx  << " " 
	   << i << "/" << MaxGenNum << endl;

      ChroBag = Jcross->apply(pop[pa1Indx],pop[pa2Indx]);

      if(ChroBag.getNumberOfElements())
      {
	 candidate = ChroBag.getElementIndex(0);

	 candidate = Jmutation->apply(candidate);

         checkShift = duplicate = 0; 
         
	 for(k=0;k<PopulationSize&&!duplicate;k++) 
	     duplicate = (*candidate==*pop[k]);

         // Make sure that it is not duplicate
	 if(!duplicate)
	 {

	    //cout << " It is not duplicate!" << endl;
            candidFV = 1.0/((SCPChromosomes*)candidate)->getCost();

	    // If it is better, bubble up is necessary
	    if( candidFV>popFV[PopulationSize-1] ) checkShift=1;

	    // delete the last element
	    delete pop[PopulationSize-1];

	    // Update the sum of fitness values
	    ftot = ftot - popFV[PopulationSize-1] + candidFV;

	    k=PopulationSize-1;
	    if( checkShift )
	    {
	        cout << " Shift needed " << endl;
		j=0;
		while( j<k && popFV[j]>candidFV) j++;

	        cout << " Shift needed " << j << endl;

		for(int m=k;m>j;m--) 
		{
		    pop[m] = pop[m-1];
		    popFV[m] = popFV[m-1];
		}
		k=m;
	    }// end of if

            // Add the new chromosome into the population
	    pop[k] = candidate;
	    popFV[k] = candidFV;
            //printPopArray(pop,PopulationSize);

            // Print some statistics
	    if( (i%10)==0) writeStatForBCGA(fileName,popFV,pop,i,PopulationSize);
	    i++;
	 }
	 else delete candidate;

      }// end of if

      if( prev_i==i ) cnt++;
      else cnt=0;

      // If it is not proucing anything, then restart some of the population
      if(cnt>20) { reInitialize(pop,popFV,PopulationSize); cnt=0; }
      prev_i=i;

   }// end of while
   stop_t = time(&dummy);
   writeStatForBCGA(fileName,popFV,pop,i,PopulationSize);

   printReport(RgaFile,bias+(stop_t-start_t),((SCPChromosomes*)pop[0]));
   printSolution(gaFile,((SCPChromosomes*)pop[0]));

   // Reclaim the memory 
   delete[] popFV;
   for(i=0;i<PopulationSize;i++)  delete pop[i];
   delete[] pop;
   delete Jmutation;
   delete Jcross;

   return;

}// end of solvewithSSGA()

/*************************************************************************/
/*************************************************************************/
void printUsage()
{
   cout << " Usage : fcover <data_file> [GREEDY|GA|RANDOM|BACK|CHV|GRD3] " 
	<< endl;
   cout << "       : <data_file> contains the input file. " << endl;
   cout << "         first line has two integer values NumberOfRows and " 
	<< endl;
   cout << "         NumberOfColumns. There should be NumberOfColumns "
	<< endl;
   cout << "         line in the file after the first line. These lines "
	<< endl;
   cout << "         contains the binary representation of the column "
	<< endl;
   cout << "         followed by the cost this column. "
	<< endl;

}// end of printUsage()

/*************************************************************************/
/*************************************************************************/

int main(int argc, char *argv[])
{   
    time_t start_t; 
    time_t stop_t; 
    time_t dummy_t; 
    File1Reader FReader ;
    SetCoverProblem *SCProblem;
    char inFileName[50];
    SCPTestDefinition test;

    // there must be a configuration file
    if(argc<2) { printUsage(); exit(1); }// end of if

    // read configuration file
    test = test.readTestDefinition(argv[1]);
    cout << test;

    srand(150);

    // Set Allocation Exception Handler
    set_new_handler(allocation_handler);

    //sprintf(inFileName,"%s",(argc>1?argv[1]:"acs294x19"));
    sprintf(inFileName,"%s",test.getDataFileName());

    // set data file
    FReader.setFileName(test.getDataFileName());

    // read data file
    switch(test.getInputFileType())
    {
       case SCPTestDefinition::CODED:
	    SCProblem = FReader.read2(); break;

       case SCPTestDefinition::ORLIB:
	    SCProblem = FReader.readORLib(); 
	    break;
       default:
	    SCProblem = FReader.read(); break;
    }// end of switch()

    // get which algorithm
    AlgorithmType aType = test.getAlgorithmType();

    if(SCProblem!=NULL) 
    {

       cout << " before buildLookUpTable() " << endl;
       SCProblem->buildLookUpTable();

       cout << " after buildLookUpTable() " << endl;
       //cout << (*SCProblem) << endl;
       //SCProblem->printLookUpTable(cout);
       cout << test << endl;

       switch(aType)
       {

	  case SCPTestDefinition::GA : // Solve with Genetic Algorithm
	       //solvewithGASTDptr(SCProblem,gaFile,RgaFile); 
	       break;

	  case SCPTestDefinition::BCGA : // Solve with Beasley Chu Genetic Algorithm
	       solvewithGASTDptrBC(SCProblem,test);
	       break;

	  case SCPTestDefinition::SSGA : // Solve with Steady State Genetic Algorithm
	       solvewithSSGA(SCProblem,test);
	       break;

          case SCPTestDefinition::BACK : // Solve with Back Greedy Choice
	       { 
		 SetCoverSolution SetCoverSolInstance(SCProblem);

		 ofstream fout("/tmp/back.out");
		 fout << *SCProblem << endl;

		 start_t = time(&dummy_t);
                 SetCoverSolInstance.solveWithGreedy(SetCoverSolution::BACK_GRDY);
		 stop_t = time(&dummy_t);

		 fout << " BACK " << endl << SetCoverSolInstance << endl;
		 fout.close();

		 printSolution( test.getBinResultFileName(), 
				&SetCoverSolInstance);

		 printReport( test.getReportFileName(), stop_t-start_t,
			      &SetCoverSolInstance);
	       }
	       break;

          case SCPTestDefinition::CHV : // Solve with Chavatal Greedy Choice
	       { 
		 SetCoverSolution SetCoverSolInstance(SCProblem);

		 ofstream fout("/tmp/chv.out");
		 fout << *SCProblem << endl;

		 start_t = time(&dummy_t);
                 SetCoverSolInstance.solveWithGreedy(SetCoverSolution::CHAVATAL_GRDY);
		 stop_t = time(&dummy_t);

		 fout << " CHAVATAL " << endl << SetCoverSolInstance << endl;
		 fout.close();

		 printSolution( test.getBinResultFileName(), 
				&SetCoverSolInstance);

		 printReport( test.getReportFileName(), stop_t-start_t,
			      &SetCoverSolInstance);
	       }
	       break;

          case SCPTestDefinition::GRD3 : // Solve with Greedy3 Greedy Choice
	       { 

		 SetCoverSolution SetCoverSolInstance(SCProblem);

		 ofstream fout("/tmp/grd3.out");
		 fout << *SCProblem << endl;

		 start_t = time(&dummy_t);
                 SetCoverSolInstance.solveWithGreedy(SetCoverSolution::GRDY_3);
		 stop_t = time(&dummy_t);

		 fout << " GREEDY #3 " << endl << SetCoverSolInstance << endl;
		 fout.close();

		 printSolution( test.getBinResultFileName(), 
				&SetCoverSolInstance);

		 printReport( test.getReportFileName(), stop_t-start_t,
		              &SetCoverSolInstance);

	       }
	       break;

          case SCPTestDefinition::RANDOM_ALG : // Random initialization 
	       {

		 SetCoverSolution SetCoverSolInstance(SCProblem);

		 ofstream fout("/tmp/rnd.out");
		 fout << *SCProblem << endl;

		 start_t = time(&dummy_t);
                 SetCoverSolInstance.solveWithRandomChoice();
		 stop_t = time(&dummy_t);

		 fout << " Random " << endl << SetCoverSolInstance << endl;
		 fout.close();

		 printSolution( test.getBinResultFileName(),
				&SetCoverSolInstance);

		 printReport( test.getReportFileName(), stop_t-start_t,
			      &SetCoverSolInstance);

	       }
	       break;

       };

       cout << test << endl;

    }// end of if 

}// end of main()

